/***********************************************************************************
* Filename 		$RCSfile: alg1_eval.c,v $
* Current CVS 		$Revision: 1.21 $
* Checked-in by 	$Author: yves $ 
*
* $Log: alg1_eval.c,v $
* Revision 1.21  2005/03/28 22:32:09  yves
* Add typedef for position
*
* Revision 1.20  2005/03/20 15:11:42  yves
* Changes before sending to CH, no checking
*
* Revision 1.19  2004/12/24 01:25:34  yves
* Opening book in binary mode implemented.
* Version 0.2.6
*
* speedup in book reading
* add open line evaluations, corrections for pawn advance and queen moves in opening
* parameter tweaking
*
* Revision 1.15  2004/12/18 13:48:49  yves
* piece position lowered to below move 20
* after this, only pawn advance counts
*
* Revision 1.14  2004/12/17 23:36:01  yves
* Serious bug fixed in the white evaluation
*
* Revision 1.13  2004/12/17 17:30:10  yves
* Add king safety to eval
*
* Revision 1.11  2004/12/16 00:18:24  yves
* Standard CVS command tags added
* FEN notation inverted, now uppercase is white and lowercase is black
*
*
***********************************************************************************/

#include "belofte.h"
#include "alg1_eval.h"
#include "alg1_board.h"
#include "alg1_moves.h"

// --------------------------------------------------------------------

static int countDoublePawn(const t_boolean bEvalForWhite);
static int whiteKingProtected(void);
static int blackKingProtected(void);
static int whitePawns(void);
static int blackPawns(void);
static int onOpenDirection(const posid iPos, const int iColour, const int iDir);
static int onOpenLine(const posid iPos, const int iColour);
static int offensiveBishop(const posid iPos, t_colour iColour);
static int offensiveRook(const posid iPos, t_colour iColour);
static int onSameDir(const posid iPos, const int iDir, const t_piece cPieceToLookFor, int *iPieces);
static t_score eval_posboard(const t_boolean isWhite);
static t_score eval_kingpos(const t_boolean isWhite);

// --------------------------------------------------------------------

extern t_alg1_board m1_board;
extern t_alg1_board m1_boardmapper;
extern t_alg1_move  m1_movelist;
extern int curply;

// --------------------------------------------------------------------

char m1_aPieceValues[BOARDARRAY_SIZE];

// --------------------------------------------------------------------
// all positions below are seen with white from bottom

const static int kingvalues[8][8] = {
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0, -2, -2, -2, -2,  0,  0 } ,
	{ 0,  3, -2, -3, -7, -4,  3,  0 } ,
	{ 0,  8,  6, -2, -1, -3,  8,  0 } 
};

const static int rookvalues[8][8] = {
	{  0,  0,  0,  0,  0,  0,  0,  0 } ,
	{  6,  8, 10, 10, 10, 10,  8,  6 } ,
	{  0,  2,  2,  2,  2,  2,  2,  0 } ,
	{  0,  0,  0,  0,  0,  0,  0,  0 } ,
	{  0,  0,  0,  0,  0,  0,  0,  0 } ,
	{  2,  0,  0,  0,  0,  1,  1, -1 } ,
	{ -5,  0,  2,  4,  4,  2,  0, -5 } ,
	{  0, -5,  1,  5,  5,  3, -5,  0 } 
};

const static int knightvalues[8][8] = {
	{ -5, -3, -1, -1, -1, -1, -3, -5 } ,
	{ -3,  0,  0,  0,  0,  0,  0, -3 } ,
	{ -1,  0,  1,  1,  1,  1,  0, -1 } ,
	{ -1,  0,  1,  1,  1,  1,  0, -1 } ,
	{ -1,  0,  1,  2,  2,  1,  0, -1 } ,
	{ -4,  0,  4,  3,  3,  4,  0, -8 } ,
	{ -3,  0,  0,  5,  5,  0,  0, -3 } ,
	{ -5, -2, -4, -4, -4, -4, -2, -5 } 
};

const static int bishopvalues[8][8] = {
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  2,  0,  0,  0,  0,  2,  0 } ,
	{ 1,  0,  5,  1,  1,  5,  0,  1 } ,
	{ 0,  1,  0,  4,  4,  0,  1,  0 } ,
	{ 0,  2,  0,  2,  3,  0,  2,  0 } ,
	{ 0,  1,  0,  0,  0,  0,  1,  0 } 
};

const static int queenvalues[8][8] = {
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0, -5 } ,
	{ 0,  0,  0,  0,  0,  0, -5,  0 } ,
	{ 0,  2, -2,  1,  1, -3,  1,  0 } ,
	{ 0,  0,  2,  5,  5,  2,  0,  0 } ,
	{ 0,  0, -2,  7,  0, -2,  0,  0 } 
};

const static int pawnvalues[8][8] = {
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } ,
	{ 0,  0,  0,  0,  0,  2,  0,  5 } ,
	{ 0,  0,  0,  2,  2,  2,  0,  0 } ,
	{ 0, -3,  3, 20, 20,  1, -3, -3 } ,
	{ 5,  0,  3,  3,  3, -5,  0,  1 } ,
	{ 4,  4,  4,-20,-20,  5,  5,  5 } ,
	{ 0,  0,  0,  0,  0,  0,  0,  0 } 
};

const static int pawnadvances[8] = {
	 0,  0,  0,  2,  10,  100,  300,  0  
};

/*----------------------------------------------------------------------+
|      Position evaluation functions                                    |
+----------------------------------------------------------------------*/

/**
 * calculate the current search depth evaluation
 */
t_score eval_pos(const t_boolean isWhite) 
{
	int iMoves;
	int iMovesWhite;
	int iMovesBlack;
	int iPos;
    posid iEpField;
	t_piece cPiece;
	struct st_alg1_movelist m1_movesbuffer;
	
	t_score fPiecePos = 0;
	t_score fWhitePiecePos = 0;
	t_score fBlackPiecePos = 0;
	
	// generate for white
	
    if (curply > MAX_GAME_LENGTH-1) {
        // if we cannot store moves anymore, consider game lost
        if (isWhite) return -SCORE_WON;
        else return SCORE_WON;
    }

	
	// calculate piece evaluation 
	for (iPos = STARTPOS; iPos < ENDPOS ; iPos++) {
		cPiece = m1_board.m_fields[iPos];
		if ((cPiece != FIELD_BORDER) && (cPiece != FIELD_EMPTY)) {
			// calculate piece evaluation 

			if (!(cPiece & (LOWERCASEBIT))) { 
				// white piece
				fWhitePiecePos += m1_aPieceValues[(int) cPiece];
			} else {
				// black piece
				fBlackPiecePos += m1_aPieceValues[(int) cPiece];
			}
		}
	}
	
	// piece values have been calculated, make score relative to total
	fPiecePos = ( fWhitePiecePos + fBlackPiecePos ) * 60 / ( fWhitePiecePos - fBlackPiecePos ) ;

	// we do not bother about possible moves if score is way off normal score
	memcpy(&m1_movesbuffer,&m1_movelist,sizeof(m1_movelist));
    iEpField = m1_board.m_epfield;
    m1_board.m_epfield = 0;

    // if row 5 is set, we know that white possibly can take ep
    if (ROW144TO64(iEpField) == 5) 
        m1_board.m_epfield = iEpField;
	iMovesWhite = fill_movelist(COLOUR_WHITE,0,NO_FILLMOVES);
 
    // if 5 was set, and it should have been 2, we need to undo the clear
    if (ROW144TO64(iEpField) == 2) 
        m1_board.m_epfield = iEpField;
    else 
        m1_board.m_epfield = 0;
	iMovesBlack = fill_movelist(COLOUR_BLACK,0,NO_FILLMOVES);
	m1_board.m_epfield = iEpField;
	memcpy(&m1_movelist,&m1_movesbuffer,sizeof(m1_movelist));
    
	// last part of evaluation, invert if needed
	// verify if mate of stalemate position
	if (isWhite) {
		if (iMovesWhite == 0) {
			if (isInCheck(myTRUE)) {
				return -SCORE_WON;
			} else {
				return SCORE_DRAW;
			}
		}
		iMoves = iMovesWhite - iMovesBlack;
	} else {
		if (iMovesBlack == 0) {
			if (isInCheck(myFALSE)) {
				return -SCORE_WON;
			} else { 
				return SCORE_DRAW;
			}
		}
		iMoves = iMovesBlack - iMovesWhite;
		
		// invert score if we are currently evaluating for black
		fPiecePos   = -fPiecePos;
	}
	
    //if (fPiecePos != 0)
    //    cmd_snd_thinking(0,fPiecePos,"Piece pos");
    //if (iMoves != 0 )
    //    cmd_snd_thinking(0,iMoves,"Moves");
        
	return fPiecePos 
		+ eval_posboard(isWhite) 
		+ ((t_score)iMoves/MOBILITY_DIVIDER)
		+ eval_kingpos(isWhite);
}

/** 
 * calculate king safety
 */
static t_score eval_kingpos(const t_boolean isWhite)
{
    t_score fKingSafety = 0;
    
    // king safety
	// casling information
	fKingSafety += ( m1_board.m_whitecastlescore - m1_board.m_blackcastlescore) * 5;
	
	// calculate malus if king has not moved after the first 10 moves
	if (curply > 20) {
		fKingSafety -= (m1_board.m_whitebegin[4] - m1_board.m_blackbegin[4]) * 2;
	}
	
	fKingSafety += whiteKingProtected();
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,-12) * 2;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,12) * 4;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,-1) * 3;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,1) * 3;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,11) * 4;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,13) * 4;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,-11) * 3;
    fKingSafety -= onOpenDirection(m1_board.m_whiteking,COLOUR_WHITE,-13) * 3;
    
	fKingSafety -= blackKingProtected();
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,-12) * 4;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,12) * 2;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,-1) * 3;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,1) * 3;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,11) * 3;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,13) * 3;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,-11) * 4;
    fKingSafety += onOpenDirection(m1_board.m_blackking,COLOUR_BLACK,-13) * 4;
    
    if (!isWhite) {
        fKingSafety = -fKingSafety;
    }
    
    //if (fKingSafety != 0)
    //    cmd_snd_thinking(0,fKingSafety,"King safety");
    
    return fKingSafety / KING_DIVIDER;
}

/**
 * return the evaluation of the board
 */
static t_score eval_posboard(const t_boolean isWhite) 
{
    t_score fBoardValue = 0;
    posid iPos;
    t_piece cPiece;
    unsigned int iWhiteAttack;
    unsigned int iBlackAttack;

    // calculate piece evaluation and boardposition values
	// also calculate protection rate of own position
	for (iPos = STARTPOS; iPos < ENDPOS ; iPos++) {
		cPiece = m1_board.m_fields[iPos];
		if ((cPiece != FIELD_BORDER) && (cPiece != FIELD_EMPTY)) {
			// calculate piece evaluation and boardposition values
            
			iWhiteAttack = isUnderAttack(iPos,myTRUE);
			iBlackAttack = isUnderAttack(iPos,myFALSE);
			if (!(cPiece & (LOWERCASEBIT))) { 
				// white piece
				if (iWhiteAttack > iBlackAttack)
					// more attacked than protected
					fBoardValue -= m1_aPieceValues[(int) cPiece] * DEVELOPMENT_FACTOR;
				if (curply<50) {
					// do not attribute special field values in the endgame
					if (cPiece == PIECE_PAWN) {
						fBoardValue += pawnvalues[7-ROW144TO64(iPos)][COL144TO64(iPos)];					
					} else if (cPiece == PIECE_KING) {
						fBoardValue += kingvalues[7-ROW144TO64(iPos)][COL144TO64(iPos)];					
					} else if (cPiece == PIECE_QUEEN) {
						fBoardValue += queenvalues[7-ROW144TO64(iPos)][COL144TO64(iPos)];					
					} else if (cPiece == PIECE_ROOK) {
						fBoardValue += rookvalues[7-ROW144TO64(iPos)][COL144TO64(iPos)];					
					} else if (cPiece == PIECE_BISHOP) {
						fBoardValue += bishopvalues[7-ROW144TO64(iPos)][COL144TO64(iPos)];					
					} else if (cPiece == PIECE_KNIGHT) {
						fBoardValue += knightvalues[7-ROW144TO64(iPos)][COL144TO64(iPos)];	
					}
				} 
				if (cPiece == PIECE_PAWN) {
					fBoardValue += pawnadvances[ROW144TO64(iPos)];
				} else if (cPiece == PIECE_ROOK) {
					fBoardValue += onOpenLine(iPos,COLOUR_WHITE) * DEVELOPMENT_FACTOR;
					fBoardValue += offensiveRook(iPos,COLOUR_WHITE);
				} else if (cPiece == PIECE_BISHOP) {
					fBoardValue += offensiveBishop(iPos,COLOUR_WHITE);
				}
			} else {
				// black piece
				if (iBlackAttack > iWhiteAttack)
					// more attacked than protected
					fBoardValue += m1_aPieceValues[(int) cPiece] * DEVELOPMENT_FACTOR;
				if (curply<50) {
					// do not attribute special field values in the endgame
					if (cPiece == BLACK_PAWN) {
						fBoardValue -= pawnvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
					} else if (cPiece == BLACK_KING) {
						fBoardValue -= kingvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
					} else if (cPiece == BLACK_QUEEN) {
						fBoardValue -= queenvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
					} else if (cPiece == BLACK_ROOK) {
						fBoardValue -= rookvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
					} else if (cPiece == BLACK_BISHOP) {
						fBoardValue -= bishopvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
					} else if (cPiece == BLACK_KNIGHT) {
						fBoardValue -= knightvalues[ROW144TO64(iPos)][COL144TO64(iPos)];
					}
				} 
				if (cPiece == BLACK_PAWN) {
					fBoardValue -= pawnadvances[7-ROW144TO64(iPos)];
				} else if (cPiece == BLACK_ROOK) {
					fBoardValue -= onOpenLine(iPos,COLOUR_BLACK) * DEVELOPMENT_FACTOR;
					fBoardValue -= offensiveRook(iPos,COLOUR_BLACK);
				} else if (cPiece == BLACK_BISHOP) {
					fBoardValue -= offensiveBishop(iPos,COLOUR_BLACK);
				}
			}
		}
	}

    // piece development, not counted as bonus points, but as malus points from the other
	// skip the queen and king in the calculation
	for (iPos=0; iPos<3; iPos++)
		fBoardValue = fBoardValue + (m1_board.m_blackbegin[iPos] - m1_board.m_whitebegin[iPos]) * DEVELOPMENT_FACTOR; 
	for (iPos=5; iPos<8; iPos++)
		fBoardValue = fBoardValue + (m1_board.m_blackbegin[iPos] - m1_board.m_whitebegin[iPos]) * DEVELOPMENT_FACTOR; 
	
	// calculate malus if queen has moved the first 8 moves
	if (curply <= 16) {
		fBoardValue -= (m1_board.m_whitebegin[3] - m1_board.m_blackbegin[3]) * DEVELOPMENT_FACTOR * (curply/2);
	}
	
	// calculate malus if rook has moved before knight or bishop
	if ((! m1_board.m_whitebegin[0]) || (! m1_board.m_whitebegin[7])) {
		fBoardValue -= ( m1_board.m_whitebegin[1] + m1_board.m_whitebegin[2] 
						 + m1_board.m_whitebegin[5] + m1_board.m_whitebegin[6] ) * DEVELOPMENT_FACTOR;
	}
	if ((! m1_board.m_blackbegin[0]) || (! m1_board.m_blackbegin[7])) {
		fBoardValue += ( m1_board.m_blackbegin[1] + m1_board.m_blackbegin[2] 
						 + m1_board.m_blackbegin[5] + m1_board.m_blackbegin[6] ) * DEVELOPMENT_FACTOR;
	}
	// calculate malus for stupid moves like Rg1, Rh2....
	if (curply<20) {
		if ((m1_board.m_fields[26] == PIECE_ROOK) || (m1_board.m_fields[33] == PIECE_ROOK))
			fBoardValue -= DEVELOPMENT_FACTOR;
		if ((m1_board.m_fields[110] == BLACK_ROOK) || (m1_board.m_fields[117] == BLACK_ROOK))
			fBoardValue += DEVELOPMENT_FACTOR;
	}		

    // add malus for double pawn
	fBoardValue += (countDoublePawn(myTRUE) - countDoublePawn(myFALSE)) * DOUBLE_PAWN_MALUS;
    
	// add pawn structure evaluation
	fBoardValue += whitePawns();
	fBoardValue -= blackPawns();
    
	// last part of evaluation, invert if needed
	if (!isWhite) {
		fBoardValue = -fBoardValue;
	}
    
    //if (fBoardValue != 0)
    //    cmd_snd_thinking(0,fBoardValue,"Board value");

    return fBoardValue / BOARD_DIVIDER;
}

/**
 * count the number of double pawns for a side to move
 */
static int countDoublePawn(const t_boolean bEvalForWhite)
{
	int iDouble = 0;
	int iCount;
	int iCol,iRow;
	t_piece cPieceToLookFor;
	
	if (bEvalForWhite) cPieceToLookFor = PIECE_PAWN;
	else cPieceToLookFor = BLACK_PAWN;
	
	for (iCol=0;iCol<=7;iCol++) {
		iCount = 0;
		for (iRow=1;iRow<7;iRow++) {
			if (m1_board.m_fields[iCol+2+(iRow+2)*12]==cPieceToLookFor) iCount++;
		}
		if (iCount>1) iDouble+= (iCount-1);
	}

	return iDouble;
}

// --------------------------------------------------------------------

static int whiteKingProtected()
{
	int iProtected = 0;
	// make sure to keep 3 pawns in line at kingside
	if ((m1_board.m_whiteking == KINGWHITESTART) && (curply<20)) {
		if ((m1_board.m_fields[43] == PIECE_PAWN) 
			&& (m1_board.m_fields[44] == PIECE_PAWN)
			&& (m1_board.m_fields[45] == PIECE_PAWN))
			iProtected += 1;
	}
	// avoid moving in the beginning g and h pawns
	if ((m1_board.m_whiteking == KINGWHITESTART) && (curply<10)) {
		if ((m1_board.m_fields[44] != PIECE_PAWN) 
			|| (m1_board.m_fields[45] != PIECE_PAWN))
			iProtected -= 2;
	}
	if (m1_board.m_whiteking == 32) {
		if ((m1_board.m_fields[43]==PIECE_PAWN) 
			&& (m1_board.m_fields[44]==PIECE_PAWN)
			&& (m1_board.m_fields[45]==PIECE_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[43]==PIECE_PAWN) 
			&& (m1_board.m_fields[44]==PIECE_PAWN)
			&& (m1_board.m_fields[57]==PIECE_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[43]==PIECE_PAWN) 
			&& (m1_board.m_fields[44]==PIECE_PAWN)
			&& (m1_board.m_fields[56]==PIECE_BISHOP)
			&& (m1_board.m_fields[45]==PIECE_PAWN)) 
			iProtected += 3;
	}
	if (m1_board.m_whiteking == 28) {
		if ((m1_board.m_fields[40]==PIECE_PAWN) 
			&& (m1_board.m_fields[39]==PIECE_PAWN)
			&& (m1_board.m_fields[38]==PIECE_PAWN)) 
			iProtected += 2;
		if ((m1_board.m_fields[40]==PIECE_PAWN) 
			&& (m1_board.m_fields[39]==PIECE_PAWN)
			&& (m1_board.m_fields[50]==PIECE_PAWN)) 
			iProtected += 2;
	}
	if (m1_board.m_whiteking == 27) {
		if ((m1_board.m_fields[40]==PIECE_PAWN) 
			&& (m1_board.m_fields[39]==PIECE_PAWN)
			&& (m1_board.m_fields[38]==PIECE_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[40]==PIECE_PAWN) 
			&& (m1_board.m_fields[39]==PIECE_PAWN)
			&& (m1_board.m_fields[50]==PIECE_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[40]==PIECE_PAWN) 
			&& (m1_board.m_fields[39]==PIECE_PAWN)
			&& (m1_board.m_fields[51]==PIECE_BISHOP)
			&& (m1_board.m_fields[38]==PIECE_PAWN)) 
			iProtected += 3;
	}

	return iProtected;
}

// --------------------------------------------------------------------

static int blackKingProtected()
{
	int iProtected = 0;
	if ((m1_board.m_blackking == KINGBLACKSTART) && (curply<20)) {
		if ((m1_board.m_fields[103] == BLACK_PAWN)
			&& (m1_board.m_fields[104] == BLACK_PAWN)
			&& (m1_board.m_fields[105] == BLACK_PAWN))
			iProtected += 1;
	}
	if ((m1_board.m_blackking == KINGBLACKSTART) && (curply<10)) {
		if ((m1_board.m_fields[104] != BLACK_PAWN) 
			|| (m1_board.m_fields[105] != BLACK_PAWN))
			iProtected -= 2;
	}
	
	if (m1_board.m_blackking == 116) {
		if ((m1_board.m_fields[103]==BLACK_PAWN) 
			&& (m1_board.m_fields[104]==BLACK_PAWN)
			&& (m1_board.m_fields[105]==BLACK_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[103]==BLACK_PAWN) 
			&& (m1_board.m_fields[104]==BLACK_PAWN)
			&& (m1_board.m_fields[93]==BLACK_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[103]==BLACK_PAWN) 
			&& (m1_board.m_fields[92]==BLACK_PAWN)
			&& (m1_board.m_fields[104]==BLACK_BISHOP)
			&& (m1_board.m_fields[105]==BLACK_PAWN)) 
			iProtected += 3;
	}
	if (m1_board.m_blackking == 112) {
		if ((m1_board.m_fields[100]==BLACK_PAWN) 
			&& (m1_board.m_fields[99]==BLACK_PAWN)
			&& (m1_board.m_fields[98]==BLACK_PAWN)) 
			iProtected += 2;
		if ((m1_board.m_fields[100]==BLACK_PAWN) 
			&& (m1_board.m_fields[99]==BLACK_PAWN)
			&& (m1_board.m_fields[86]==BLACK_PAWN)) 
			iProtected += 2;
	}
	if (m1_board.m_blackking == 111) {
		if ((m1_board.m_fields[100]==BLACK_PAWN) 
			&& (m1_board.m_fields[99]==BLACK_PAWN)
			&& (m1_board.m_fields[98]==BLACK_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[100]==BLACK_PAWN) 
			&& (m1_board.m_fields[99]==BLACK_PAWN)
			&& (m1_board.m_fields[86]==BLACK_PAWN)) 
			iProtected += 3;
		if ((m1_board.m_fields[100]==BLACK_PAWN) 
			&& (m1_board.m_fields[87]==BLACK_PAWN)
			&& (m1_board.m_fields[99]==BLACK_BISHOP)
			&& (m1_board.m_fields[98]==BLACK_PAWN)) 
			iProtected += 3;
	}
	
	return iProtected;
}

// --------------------------------------------------------------------

static int whitePawns(void)
{
	int iResult = 0;
	int iPos;
	
	for (iPos=STARTPOS;iPos<ENDPOS;iPos++) {
		if (m1_board.m_fields[iPos] == PIECE_PAWN) {
			// are pawns being protected?
			if ( (m1_board.m_fields[iPos-11] == PIECE_PAWN) 
				 || (m1_board.m_fields[iPos-13] == PIECE_PAWN))
				iResult+=2;
			// are pawns forming a wall?
			if ( (m1_board.m_fields[iPos-1] == PIECE_PAWN) 
				 || (m1_board.m_fields[iPos+2] == PIECE_PAWN))
				iResult++;
		}
	}
	
	// check if d3 or e3 are blocking the pawn
	if ((m1_board.m_fields[41] == PIECE_PAWN) 
		&& (m1_board.m_fields[53] != FIELD_EMPTY)) iResult -=60;

	if ((m1_board.m_fields[42] == PIECE_PAWN) 
		&& (m1_board.m_fields[54] != FIELD_EMPTY)) iResult -=60;

	return iResult;
}

// --------------------------------------------------------------------

static int blackPawns(void)
{
	int iResult = 0;
	int iPos;
	
	for (iPos=STARTPOS;iPos<ENDPOS;iPos++) {
		if (m1_board.m_fields[iPos] == BLACK_PAWN) {
			// are pawns being protected?
			if ( (m1_board.m_fields[iPos+11] == BLACK_PAWN) 
				 || (m1_board.m_fields[iPos+13] == BLACK_PAWN))
				iResult+=2;
			// are pawns forming a wall?
			if ( (m1_board.m_fields[iPos-1] == BLACK_PAWN) 
				 || (m1_board.m_fields[iPos+2] == BLACK_PAWN))
				iResult++;
		}
	}

	// check if d6 or e6 are blocking the pawn
	if ((m1_board.m_fields[101] == BLACK_PAWN) 
		&& (m1_board.m_fields[89] != FIELD_EMPTY)) iResult -=60;
	if ((m1_board.m_fields[102] == BLACK_PAWN) 
		&& (m1_board.m_fields[90] != FIELD_EMPTY)) iResult -=60;

	return iResult;
}

/**
 * check if rook is on open line
 */
static int onOpenLine(const posid iPos,const int iColour)
{
	int iResult = 3;
	int i;
	
	if (iColour==COLOUR_WHITE) {
		for (i = iPos + 12;i<ENDPOS; i+= 12) {
			if (m1_board.m_fields[i] == PIECE_PAWN) {
				iResult = 0;
				break;
			} else if (m1_board.m_fields[i] == BLACK_PAWN) {
				iResult = 1;
				break;
			}
		}
	} else {
		for (i = iPos - 12;i>=STARTPOS; i-= 12) {
			if (m1_board.m_fields[i] == BLACK_PAWN) {
				iResult = 0;
				break;
			} else if (m1_board.m_fields[i] == PIECE_PAWN) {
				iResult = 1;
				break;
			}
		}
	}
	
	return iResult;
}

/**
 * return 2 if we are seeing a first enemy piece in a certain direction
 * return 1 if we are seeing a border after 2 empty cases
 * return 0 in case of own piece or border
 */
static int onOpenDirection(const posid iPos, const int iColour, const int iDir)
{
    int iResult = 0;
    int iIterator = 0;
    t_piece cPiece;
    
    posid iDest = iPos + iDir;
    cPiece = m1_board.m_fields[iDest];
    while (cPiece != FIELD_BORDER) {
        if (cPiece != FIELD_EMPTY) {
            if (iColour == COLOUR_WHITE) {
                if (cPiece == BLACK_KING) {
                    iResult = 1;
                } else if (cPiece && LOWERCASEBIT) {
                    // we are white and we are encountering a black piece
                    iResult = 2;
                }
            } else {
                if (cPiece == PIECE_KING) {
                    iResult = 1;
                } else if (!(cPiece && LOWERCASEBIT)) {
                    // we are black and are encountering a white piece
                    iResult = 2;
                }
            }
            break;
        }
        iDest += iDir;
        cPiece = m1_board.m_fields[iDest];
        iIterator++;
    }
    
    if ((iIterator>1) && (cPiece == FIELD_BORDER))
        iResult = 1;

    return iResult;
}

/**
 *  see if Bishop is on same line as other higher pieces
 */
static int offensiveBishop(const posid iPos, enum tColour iColour)
{
	t_score fResult = 0.0;
	int iPieces;
	
	if (iColour == COLOUR_WHITE) {
		if (onSameDir(iPos,13,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,11,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-13,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-11,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;

		if (onSameDir(iPos,13,BLACK_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		else if (onSameDir(iPos,11,BLACK_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		else if (onSameDir(iPos,-13,BLACK_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		else if (onSameDir(iPos,-11,BLACK_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;

		if (onSameDir(iPos,13,BLACK_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
		else if (onSameDir(iPos,11,BLACK_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
		else if (onSameDir(iPos,-13,BLACK_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
		else if (onSameDir(iPos,-11,BLACK_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
	} else {
		if (onSameDir(iPos,13,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,11,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-13,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-11,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		
		if (onSameDir(iPos,13,PIECE_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		else if (onSameDir(iPos,11,PIECE_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		else if (onSameDir(iPos,-13,PIECE_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		else if (onSameDir(iPos,-11,PIECE_QUEEN,&iPieces))
			fResult += 8 / (t_score) iPieces;
		
		if (onSameDir(iPos,13,PIECE_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
		else if (onSameDir(iPos,11,PIECE_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
		else if (onSameDir(iPos,-13,PIECE_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
		else if (onSameDir(iPos,-11,PIECE_ROOK,&iPieces))
			fResult += 3 / (t_score) iPieces;
	}
	
	return (int) fResult;
}

// --------------------------------------------------------------------

static int offensiveRook(const posid iPos, t_colour iColour)
{
	t_score fResult = 0.0;
	int iPieces;
	
	if (iColour == COLOUR_WHITE) {
		if (onSameDir(iPos,12,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,1,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-12,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-1,BLACK_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		
		if (onSameDir(iPos,12,BLACK_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		else if (onSameDir(iPos,1,BLACK_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		else if (onSameDir(iPos,-12,BLACK_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		else if (onSameDir(iPos,-1,BLACK_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		
	} else {
		if (onSameDir(iPos,12,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,1,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-12,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		else if (onSameDir(iPos,-1,PIECE_KING,&iPieces))
			fResult += 10 / (t_score) iPieces;
		
		if (onSameDir(iPos,12,PIECE_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		else if (onSameDir(iPos,1,PIECE_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		else if (onSameDir(iPos,-12,PIECE_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		else if (onSameDir(iPos,-1,PIECE_QUEEN,&iPieces))
			fResult += 6 / (t_score) iPieces;
		
	}
	
	return (int) fResult;
}

// --------------------------------------------------------------------

static int onSameDir(const posid iPos, const int iDir, const t_piece cPieceToLookFor, int *iPieces)
{
	t_piece cPiece;
	posid iDest = iPos + iDir;
	int iHit = myFALSE;
	
	*iPieces=1;
	
	while (1) {
		cPiece = m1_board.m_fields[iDest];
		if (cPiece == FIELD_BORDER) 
			break;
		if (cPiece != FIELD_EMPTY) 
			*iPieces++;
		if (cPiece == cPieceToLookFor) {
			iHit = myTRUE;
			break;
		}
		iDest += iDir;
	}

	// comment out, avoid division by zero, calculation in calling function should be adapted though for
	// better evaluation
	// if (iHit) *iPieces--; 
	
	return iHit;
}

/**
 * generate internal buffers to speed up evaluation functions
 */
void fill_moveevalbuffers(void)
{
	int i;
	
	for (i=0;i<BOARDARRAY_SIZE;i++) m1_aPieceValues[i] = 0;
	
	m1_aPieceValues[PIECE_PAWN] = 1;
	m1_aPieceValues[BLACK_PAWN] = -1;
	m1_aPieceValues[PIECE_KNIGHT] = 3;
	m1_aPieceValues[BLACK_KNIGHT] = -3;
	m1_aPieceValues[PIECE_BISHOP] = 3;
	m1_aPieceValues[BLACK_BISHOP] = -3;
	m1_aPieceValues[PIECE_ROOK] = 5;
	m1_aPieceValues[BLACK_ROOK] = -5;
	m1_aPieceValues[PIECE_QUEEN] = 10;
	m1_aPieceValues[BLACK_QUEEN] = -10;
	m1_aPieceValues[PIECE_KING] = 0;
	m1_aPieceValues[BLACK_KING] = 0;
	m1_aPieceValues[FIELD_EMPTY] = 0;
}

// --------------------------------------------------------------------
// eof