/***********************************************************************************
* Filename 		$RCSfile: board.c,v $
* Current CVS 		$Revision: 1.23 $
* Checked-in by 	$Author: yves $ 
*
* $Log: board.c,v $
* Revision 1.23  2005/03/28 22:28:12  yves
* Add typedef for position
*
* Revision 1.22  2004/12/24 03:22:47  yves
* Correction on ambiguous capture moves
*
* Revision 1.21  2004/12/24 01:25:34  yves
* Opening book in binary mode implemented.
* Version 0.2.6
*
* Revision 1.20  2004/12/21 00:43:10  yves
* opening book implemented
*
* Revision 1.19  2004/12/18 15:35:27  yves
* add input in sanmove
*
* Revision 1.18  2004/12/18 13:46:57  yves
* Flags for castling corrected and reduced to 2, one for each side
*
* Revision 1.17  2004/12/18 11:53:11  yves
* add ep move
*
* Revision 1.16  2004/12/18 02:30:58  yves
* When double moves exist, make sure to write the correct move
*
* Revision 1.15  2004/12/17 23:41:52  yves
* fen string correct '1' entry, correct black to move interpretation
*
* Revision 1.14  2004/12/17 17:30:56  yves
* Full fen string loading, fen string storing
*
* Revision 1.13  2004/12/16 03:03:31  yves
* Add SAN moves in history, will be used for opening book handling
*
* Revision 1.12  2004/12/16 00:18:25  yves
* Standard CVS command tags added
* FEN notation inverted, now uppercase is white and lowercase is black
*
*
***********************************************************************************/

#include "belofte.h"
#include "board.h"
#include "alg1_search.h"
#include "alg1_moves.h"
#include "alg1_board.h"

// --------------------------------------------------------------------

int curply;

t_board m_board;

struct st_history m_boardhistory[MAX_GAME_LENGTH];

// --------------------------------------------------------------------

extern t_alg1_moves m1_movelist;
extern int busy_book_creating;

// --------------------------------------------------------------------

static void board_setpiece(const posid iIndex, const t_piece cPiece);
static void board_setinvpiece(const posid iIndex, const t_piece cPiece);
static int board_getpiece(const posid iIndex);

static void board_sanmove(char *szDest, const char *szMove);
static char board_sanmoveaddcheck(void);
static void board_sanambigious(char *szDest, const t_piece cPiece, const posid iFrom, const posid iTo);
static void board_fenstring(char *szDest, const char *szCoordMove);

static void add_char(char *szDest, const char cChar);
static void ins_char(char *szDest, const char cChar, const int iPos);

// board layout is left bottum index = 0, right bottom = 7, left top = 56, right top = 63
// board element is m_field[column][row]

// --------------------------------------------------------------------

void board_print(void)
{
	int iRow,iColumn;
	int c;
    char szBuffer[MAX_DATANAME_LENGTH];
    
    if (!busy_book_creating) {
        // do not print the board position during book creation, speedup
        // for creating book faster
        sprintf(szBuffer,"\n FEN: %s \n",m_boardhistory[min(curply,MAX_GAME_LENGTH-1)].szFEN);
        cmd_snd(szBuffer);

        for(iRow=0;iRow<8;iRow++) {
            cmd_snd("+---+---+---+---+---+---+---+---+");
            for(iColumn=0;iColumn<8;iColumn++) {
                c = m_board.m_fields[iColumn][7-iRow];
                printf("| %c ",c?(char)c:' ');
            }
            cmd_snd("|");
        }
        cmd_snd("+---+---+---+---+---+---+---+---+");
    }
}

/**
 * apply standard move
 * move entering is a move in the format from-case, to case, attribute
 * will return a code if there is something special about the move applied
 */
int board_applymove(const char *szCoordMove)
{
	posid iFrom;
	posid iTo;
	t_piece cPieceMoved;
    t_piece cPieceTo;
    char cCheck;
	int cFlag = FIELD_EMPTY;
	
	// keep past board for undo
	if (curply<MAX_GAME_LENGTH-1) {
		memcpy(&(m_boardhistory[curply].m_board),&m_board,sizeof(m_board));
		strcpy(m_boardhistory[curply].szIntMove,szCoordMove);
		board_sanmove(m_boardhistory[curply].szSANMove,szCoordMove);
	}
	
	if (strlen(szCoordMove) > 3) {
		// very shorthand way to copy move into local working variables
		iFrom = (szCoordMove[0] - 'a') + (szCoordMove[1] - '1') * 8;
		iTo   = (szCoordMove[2] - 'a') + (szCoordMove[3] - '1') * 8;
	
		// get a flag for promotion, ep, check, ....
		if (strlen(szCoordMove)>4) cFlag = szCoordMove[4];
	
		cPieceMoved = board_getpiece(iFrom);
		
		// add ep move, if not double move and switch columns
		if ((cPieceMoved == PIECE_PAWN) 
			&& ((iTo-iFrom)!=24)
			&& (board_getpiece(iTo)==FIELD_EMPTY)
			&& (M1_LIJN(iFrom) != M1_LIJN(iTo))) {
			board_setpiece((posid)(iTo-8),FIELD_EMPTY);
		}
		if ((cPieceMoved == BLACK_PAWN) 
			&& ((iFrom-iTo)!=24)
			&& (board_getpiece(iTo)==FIELD_EMPTY)
			&& (M1_LIJN(iFrom) != M1_LIJN(iTo))) {
			board_setpiece((posid)(iTo+8),FIELD_EMPTY);
		}		

        cPieceTo = board_getpiece(iTo);
		if ((cPieceTo != FIELD_EMPTY)
			|| (cPieceMoved == PIECE_PAWN)
			|| (cPieceMoved == BLACK_PAWN)) {
			// capture, pawn move, increase counter
			m_board.m_rule50 = 0;
		} else {
			// increase 50 moves rule counter
			m_board.m_rule50++;
		}
			
		board_setpiece(iTo,cPieceMoved);
		board_setpiece(iFrom,FIELD_EMPTY);
	
		if ((cPieceMoved == PIECE_KING) || (cPieceMoved == BLACK_KING)) {
			// in case of a king move, evaluate if there is no castling to be done
			if (abs(iFrom-iTo)==2) {
				// castling move, determine side
				if (iTo == 2) {
					board_setpiece(0,FIELD_EMPTY);
					board_setpiece(3,PIECE_ROOK);
				} else if (iTo == 6 ) {
					board_setpiece(7,FIELD_EMPTY);
					board_setpiece(5,PIECE_ROOK);
				} else if (iTo == 58) {
					board_setpiece(56,FIELD_EMPTY);
					board_setpiece(59,BLACK_ROOK);
				} else {
					board_setpiece(63,FIELD_EMPTY);
					board_setpiece(61,BLACK_ROOK);
				}
			}
			// clear castling flags
			if (cPieceMoved == PIECE_KING) {
				m_board.m_piecemoved[COLOUR_WHITE][0] = myTRUE;
				m_board.m_piecemoved[COLOUR_WHITE][1] = myTRUE;
			} else {
				m_board.m_piecemoved[COLOUR_BLACK][0] = myTRUE;
				m_board.m_piecemoved[COLOUR_BLACK][1] = myTRUE;
			}
		}
		
		// clear castling flags
		if (cPieceMoved == PIECE_ROOK) {
			if (iFrom == 0) m_board.m_piecemoved[COLOUR_WHITE][0] = myTRUE;
			if (iFrom == 7) m_board.m_piecemoved[COLOUR_WHITE][1] = myTRUE;
		}

		if (cPieceMoved == BLACK_ROOK) {
			if (iFrom == 56) m_board.m_piecemoved[COLOUR_BLACK][0] = myTRUE;
			if (iFrom == 63) m_board.m_piecemoved[COLOUR_BLACK][1] = myTRUE;
		}
						
		// also check promotion and ep moves
		if ((cFlag==PROMOTION_QUEEN) || (cFlag==PROMOTION_ROOK) || (cFlag==PROMOTION_BISHOP) || (cFlag==PROMOTION_KNIGHT)) {
			// promotion moves are always stored in uppercase
			board_setpiece(iTo,cPieceMoved<'a'?cFlag:tolower(cFlag));		
		}

		// and set the board ply
		curply++;
		m_board.m_ply = curply;
		
		search_apply_move(szCoordMove);		
		
		if (curply<MAX_GAME_LENGTH-1) {
			cCheck = board_sanmoveaddcheck();
            if (cCheck != ' ') {
                int iLen = strlen(m_boardhistory[curply-1].szSANMove);
                m_boardhistory[curply-1].szSANMove[iLen] = cCheck;
                m_boardhistory[curply-1].szSANMove[iLen+1] = '\0';
            }
			board_fenstring(m_boardhistory[curply].szFEN, szCoordMove);
		}
	}
	
	// TODO: some handlers about special moves, flags returning from the engine about theoretic draws, etc
	// for now, we will just try to count a lack of material to force a gameend
		
	return board_getmetadata(szCoordMove);
}

/**
 * apply move in san format
 * for this we first have to make a list of all possible moves and then map 
 * the san move to this list
 */
int board_sanapplymove(char *szMove,char *szCoordMove)
{
	int i;
	int iMoves;
	t_alg1_moves m1_movesbuffer;
	
	strcpy(szCoordMove,"");
	memcpy(&m1_movesbuffer,&m1_movelist,sizeof(m1_movelist));
	iMoves = fill_movelist(curply%2?COLOUR_BLACK:COLOUR_WHITE, myFALSE, FILLMOVES);
	normalize_movelist();
	for (i=0;i<iMoves;i++) {
		search_printablemove(i,szCoordMove);
		board_applymove(szCoordMove);
		if (strcmp(m_boardhistory[curply-1].szSANMove,szMove)==0) {
			// there is a match, so this move can be applied
			board_unapplymove();
			break;
		}
		board_unapplymove();
		strcpy(szCoordMove,"");
	}
	// copy back
	memcpy(&m1_movelist,&m1_movesbuffer,sizeof(m1_movelist));
	
	// if there was a match, the szIntMove has been retained
	if (strlen(szCoordMove)>0) {
		board_applymove(szCoordMove);
		return MOVERESULT_OK;
	}
	
	return MOVERESULT_NIL;  // no matching move, nothing done
}

/**
 * unapply standard move
 * return false if unapply did not succeed
 */
int board_unapplymove(void)
{
	// keep past board for undo
	if ((curply<MAX_GAME_LENGTH-1) && (curply>0)) {
		curply--;
		m_board.m_ply = curply;
		memcpy(&m_board,&(m_boardhistory[curply].m_board),sizeof(m_board));
		
		// bring search engine up to date
		search_initboard();
		
		return 1;
	}
	
	return 0;
}

/**
 * get some statistics about the board, to test threefold position, ...
 * return 0: no board statistics of any meaning
 * return 1: draw by lack of material
 */
int board_getmetadata(const char *szFlag)
{
	posid i;
	int iPieces=0;
	
	if (strlen(szFlag) == 2) {
		// we have some flag returning from the search engine, 
		// return it back after conversion from string
		return -(szFlag[1] - '0');
	}
	
	for (i=0;i<64;i++)
		if (board_getpiece(i) != FIELD_EMPTY) iPieces++;
	
	if (iPieces<3) 
		return 1;  // draw by lack of material
	
	return 0;  // nothing applied
}

/**
 * setboard function, will initialise by default to this
 * cmd_setboard("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
 */
void board_setup(char *szFen)
{
	int i,j;
	int iFenLength,iPosLength,iIndex;
	char *szPly;
	
	// empty the board first
	for(i=0;i<8;i++)
		for(j=0;j<8;j++)
			m_board.m_fields[i][j] = FIELD_EMPTY;
	
	iFenLength = strlen(szFen);
	iPosLength = iFenLength;

	// skip blanks at start of position
	while (szFen[0]==' ') szFen++;
	
	// calculate where the setting for castling are, they are separated by a space
	for(i=0;i<iFenLength;i++)
		if (szFen[i]==' ') { 
			iPosLength = i; break; 
		}
			
	iIndex=0;
	for(i=0;i<iPosLength;i++) {
		switch (szFen[i]) {
			case PIECE_PAWN: case PIECE_ROOK: case PIECE_KNIGHT: case PIECE_BISHOP: case PIECE_QUEEN: case PIECE_KING:
			case BLACK_PAWN: case BLACK_ROOK: case BLACK_KNIGHT: case BLACK_BISHOP: case BLACK_QUEEN: case BLACK_KING:
				board_setinvpiece((posid) iIndex, (t_piece) szFen[i]);
				iIndex++;
				break;
			case '/':  // go beginning to next row
				iIndex = (iIndex + 7) & 0xF8 ;
				break;
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8': // increment by number of empty cases
				iIndex += (szFen[i] - '0');
				// if skip to new row, wait until a slash forces to do so
				if (iIndex == (iIndex & 0xF8 ) ) iIndex--;
		}
	}
	
	// we have done all pieces, now store side to move information
	m_board.m_sidetomove = szFen[iPosLength+1] == 'w' ? COLOUR_WHITE : COLOUR_BLACK;
			
	// read castling information, first reset flags
	for (i=0;i<2;i++) {
		m_board.m_piecemoved[COLOUR_WHITE][i] = myTRUE;
		m_board.m_piecemoved[COLOUR_BLACK][i] = myTRUE;
	}
	iIndex = iPosLength+3;
	while (szFen[iIndex]!=' ') {
		// we actually only flag as the rook having moved, it is enough to avoid castling
		if (szFen[iIndex]=='K') m_board.m_piecemoved[COLOUR_WHITE][1] = myFALSE;
		if (szFen[iIndex]=='Q') m_board.m_piecemoved[COLOUR_WHITE][0] = myFALSE;
		if (szFen[iIndex]=='k') m_board.m_piecemoved[COLOUR_BLACK][1] = myFALSE;
		if (szFen[iIndex]=='q') m_board.m_piecemoved[COLOUR_BLACK][0] = myFALSE;
		iIndex++;
	}
	
	// read ep information
	iIndex++; // we are now on the ep reading
	if (szFen[iIndex] != '-') {
		m_board.m_epfield = (posid) (szFen[iIndex] - 'a' + ((szFen[iIndex+1] - '1') * 8));
		iIndex++;
	} else {
        m_board.m_epfield = 0;
    }
	
	iIndex+=2; // we are now on the 50 rules reading
	szPly = szFen+iIndex;

	// read 50 move rule information
	// read current ply information
	sscanf(szPly,"%d %d",&m_board.m_rule50,&curply);
	curply = (curply * 2) - 2;
	if (m_board.m_sidetomove == COLOUR_BLACK) curply++;
	m_board.m_ply = curply;

	// store the initial position fen string
	board_fenstring(m_boardhistory[curply].szFEN, "");

	// store for prefilling the history, after the first move, this board will be overwritten
	// anyway but other operations might rely on the board being initialised
	memcpy(&(m_boardhistory[curply].m_board),&m_board,sizeof(m_board));
	
	// copy the board into the search engine
	search_init();
	
	return;
}

/**
 * copy the fen string into the dest
 */
static void board_fenstring(char *szDest, const char *szMove)
{
	int iRow, iColumn;
	int iEmpty;
	t_piece cPiece;
	char szCurPly[MAX_FEN_LENGTH];
	
	strcpy(szDest,"");
	
	// board values
	for(iRow=0;iRow<8;iRow++) {
		iEmpty = 0;
		for(iColumn=0;iColumn<8;iColumn++) {
			cPiece = m_board.m_fields[iColumn][7-iRow];
			if (cPiece == FIELD_EMPTY) {
				iEmpty++;
			} else {
				if (iEmpty) add_char(szDest,(const char)('0'+iEmpty));
				add_char(szDest,(const char) cPiece);
				iEmpty = 0;
			}
		}
		if (iEmpty) add_char(szDest,(const char)('0'+iEmpty));
		if (iRow<7) add_char(szDest,'/');
	}
	
	// side to move
	add_char(szDest,' ');
	add_char(szDest,(const char)(curply%2?'b':'w'));
	
	// castling flags
	add_char(szDest,' ');
	if (! m_board.m_piecemoved[COLOUR_WHITE][1]) add_char(szDest,'K');
	if (! m_board.m_piecemoved[COLOUR_WHITE][0]) add_char(szDest,'Q');
	if (! m_board.m_piecemoved[COLOUR_BLACK][1]) add_char(szDest,'k');
	if (! m_board.m_piecemoved[COLOUR_BLACK][0]) add_char(szDest,'q');
	
	// if none of the flags added, we add a dash
	if ((szDest[strlen(szDest)-1] == ' ')) add_char(szDest,'-');

	add_char(szDest,' ');

	// ep field
	if (strlen(szMove)>3) {
		posid iFrom = (posid) ((szMove[0] - 'a') + (szMove[1] - '1') * 8);
		posid iTo   = (posid) ((szMove[2] - 'a') + (szMove[3] - '1') * 8);
		
		if (board_getpiece(iTo) == PIECE_PAWN) {
			if ((iTo - iFrom) == 16) {
				add_char(szDest,szMove[0]);
				add_char(szDest,szMove[1]+1);
			} else {
				add_char(szDest,'-');
			}
		} else if (board_getpiece(iTo) == BLACK_PAWN) {
			if ((iFrom - iTo) == 16) {
				add_char(szDest,szMove[0]);
				add_char(szDest,szMove[3]+1);
			} else {
				add_char(szDest,'-');
			}
		} else {
			add_char(szDest,'-');
		}
	} else {
		// no valid move given, so no ep move possible
		add_char(szDest,'-');
	}
	
	// 50 move rule and ply
	sprintf(szCurPly," %d %d",m_board.m_rule50,((curply+2)/2));
	strcat(szDest,szCurPly);
}

/**
 * normalize sanmove in current position from szMove as listed
 */
static void board_sanmove(char *szDest,const char *szMove)
{
	// TODO: add from row and/or column if duplicate interpretation is possible
	// normally this is done by looking at all possible moves and see
	// if there is the same piece that can go to the same destination
	// beware: exception is when both rown and column need to be specified
	
	posid iFrom;
	posid iTo;
	t_piece cPiece;
	
	strcpy(szDest,"");
	
	if (strlen(szMove) > 3) {
		// very shorthand way to copy move into local working variables
		iFrom = (posid) ((szMove[0] - 'a') + (szMove[1] - '1') * 8);
		iTo   = (posid) ((szMove[2] - 'a') + (szMove[3] - '1') * 8);
		
		cPiece = board_getpiece(iFrom);
		
		if ((cPiece == PIECE_KING) || (cPiece == BLACK_KING)) {
			// look for castling
			if (abs(iFrom-iTo)==2) {
				// only now we have a castling move, so load it
				if (iFrom > iTo) {
					// long castle
					strcpy(szDest,"O-O-O");
				} else { 
					// short castle
					strcpy(szDest,"O-O");
				}
				return;
			}
		}
		
		if ((cPiece == PIECE_PAWN) || (cPiece == BLACK_PAWN)) {
			if ((iFrom % 8) != (iTo % 8)) {
				// pawn captures are actually noted with the from field and a 'x' sign
				add_char(szDest,szMove[0]);
				add_char(szDest,'x');
			}
		} else {
			add_char(szDest,toupper(cPiece));
			if (board_getpiece(iTo) != FIELD_EMPTY)
				add_char(szDest,'x');
		}
		
		// destination field
		add_char(szDest,szMove[2]);
		add_char(szDest,szMove[3]);
		
		if ((toupper(cPiece) == PIECE_QUEEN) 
			|| (toupper(cPiece) == PIECE_ROOK)
			|| (toupper(cPiece) == PIECE_BISHOP)
			|| (toupper(cPiece) == PIECE_KNIGHT)) {
			// check if another piece could have made this move
			board_sanambigious(szDest,cPiece,iFrom,iTo);
		}
			
		
		// add promotion flags
		if ((cPiece == PIECE_PAWN) && (szMove[3]=='8')) {
			add_char(szDest,'=');
			add_char(szDest,szMove[4]);
		}

		if ((cPiece == BLACK_PAWN) && (szMove[3]=='1')) {
			add_char(szDest,'=');
			add_char(szDest,szMove[4]);
		}
	}
}

/**
 * add a plus sign to the move if the piece is in check
 * add a hash in case of mate
 */
static char board_sanmoveaddcheck(void)
{
	// add + or # sign accordingly
	
	if (fill_movelist(curply%2?COLOUR_BLACK:COLOUR_WHITE,myFALSE,myFALSE) < 1) {
		return '#';
	}
	
	if (isInCheck((curply%2)==0))
        return '+';
    
    return ' ';
}

/**
 * check if there are 2 different pieces that can go to the destination
 * if so, change the move to make it more descriptive
 */
static void board_sanambigious(char *szDest, const t_piece cPiece, const posid iFrom, const posid iTo)
{
	posid iPos;
	int iMatches = 0;
	int iMoves;
	int i;
	t_alg1_moves m1_movesbuffer;
	
	for (iPos=0;iPos<64;iPos++) {
		if (iPos != iFrom) {
			if (board_getpiece(iPos)==cPiece) {
				// there is a second piece, check if it can play to the given position
				iMatches=1;
			}
		}
	}
	
	if (iMatches==0) return;
	
	// we know that there is at least a second piece of the same kind
	// create list of all possible moves for later calculation

	memcpy(&m1_movesbuffer,&m1_movelist,sizeof(m1_movelist));
	iMoves = fill_movelist(curply%2?COLOUR_BLACK:COLOUR_WHITE, myFALSE, FILLMOVES);
	normalize_movelist();
	
	// we loop through all, because there could be multiple pieces moving to the same
	// theoretically anyway, because normally there is only conflict after promotion
	for (iPos=0;iPos<64;iPos++) {
		if (iPos != iFrom) {
			if (board_getpiece(iPos)==cPiece) {
				// there is a second piece, check if it can play to the given position
				for (i=0;i<iMoves;i++) {
					if ((m1_movelist.m1_moves[i].ucFrom == iPos) 
						&& (m1_movelist.m1_moves[i].ucTo == iTo)) {
						// a move from this position is possible to the destination
						if (M1_LIJN(m1_movelist.m1_moves[i].ucFrom) == M1_LIJN(iPos)) {
							// add col, it will be enough to make it different
							ins_char(szDest,M1_LIJN(iFrom)+'a',strlen(szDest)-(szDest[1]=='x'?3:2));
						} else if (M1_RIJ(m1_movelist.m1_moves[i].ucFrom) == M1_RIJ(iPos)) {
							// add row, it will be enough to make it different
							ins_char(szDest,M1_RIJ(iFrom)+'1',strlen(szDest)-(szDest[1]=='x'?3:2));
						}
					}
				}
			}
		}
	}
	
	// copy back
	memcpy(&m1_movelist,&m1_movesbuffer,sizeof(m1_movelist));
}

// --------------------------------------------------------------------

static void board_setpiece(const posid iIndex, const t_piece cPiece)
{
	m_board.m_fields[M1_LIJN(iIndex)][M1_RIJ(iIndex)] = cPiece;
}

/**
 * because the fen string goes row 8 to row 1, we invert the positioning here
 * only used for fen reading
 */
static void board_setinvpiece(const posid iIndex, const t_piece cPiece)
{
	m_board.m_fields[M1_LIJN(iIndex)][7-M1_RIJ(iIndex)] = cPiece;
}

// --------------------------------------------------------------------

static int board_getpiece(const posid iIndex)
{
	return m_board.m_fields[M1_LIJN(iIndex)][M1_RIJ(iIndex)];
}

// --------------------------------------------------------------------

static void add_char(char *szDest, const char cChar)
{
	int iLen = strlen(szDest);
	
	szDest[iLen] = cChar;
	szDest[iLen+1] = '\0';
}

/**
 * insert a single character at position iPos
 */
static void ins_char(char *szDest, const char cChar, const int iPos)
{
	char szBuffer[MAX_DATANAME_LENGTH];
	int i;
	
	for (i=0;i<iPos;i++)
		szBuffer[i] = szDest[i];
	szBuffer[iPos]=cChar;  
	szBuffer[iPos+1]='\0';
	strcat(szBuffer,szDest+i);
	
	strcpy(szDest,szBuffer);
}

// --------------------------------------------------------------------
// eof