/***********************************************************************************
* Filename			$RCSfile: usercmd.c,v $
* Current CVS 		$Revision: 1.33 $
* Checked-in by 	$Author: yves $ 
*
* Revision 1.29  2004/12/19 15:47:23  yves
* add save command, autosave when result is given
* add smove command, input in san format
* about command added
*
* Revision 1.25  2004/12/17 17:34:41  yves
* add time interpreting
*
* Revision 1.24  2004/12/16 03:05:15  yves
* add game command to print moves played
*
* Revision 1.23  2004/12/16 00:18:25  yves
* Standard CVS command tags added
* FEN notation inverted, now uppercase is white and lowercase is black
*
***********************************************************************************/

// --------------------------------------------------------------------

// required for select
#include <sys/types.h>

#if defined(_MSC_VER)

#pragma warning( disable : 4100 )
#define UNUSED
#include <io.h>
#include <Winsock2.h>

#else

#include <sys/time.h>
#include <unistd.h>

// required for getch
#include <curses.h>

#define UNUSED __attribute__((unused))
#endif 

#include "belofte.h"
#include "board.h"
#include "alg1_book.h"
#include "alg1_eval.h"

// --------------------------------------------------------------------
// data structures needed for commands

int xboard_mode;	// no prompt, xboard compatible interface if set
int random_mode;	// randomize game if set
int force_mode;		// engine completely on halt if set
int ponder_mode;	// do ponder during client thinking
int level_mode;		// define what kind of level is taken
int post_mode;		// send output while thinking
int interface_mode; // interface towards xboard set
int processing_mode; // set if no display is to be given on cmd_new
int ics_mode;		// set if on chess server
int time_left;		// number of seconds left
int thinking;		// actually pondering or thinking at the moment

int algoritme;
int computer_engines[2] = { 0, 1 }; // default to computer engine for black
int level;			// the level to play at

t_depth maxdepth;

char szOpponentName[MAX_DATANAME_LENGTH];
char szResult[MAX_DATANAME_LENGTH];
char szSite[MAX_DATANAME_LENGTH];

// forward declaration

struct st_command {
	char *cmd_name;
	void (*cmd_ptr)(char*,char*);
	char *cmd_help;
	int cmd_interfacelevel:3;					// at what interface level this command can be used
	int cmd_paramdata;						// give more detail about mandatory or optional parameters
};

static struct st_command list_commands[100];

// --------------------------------------------------------------------

extern int curply;			// current movenumber

extern struct st_history m_boardhistory[MAX_GAME_LENGTH]; // for game history

/*----------------------------------------------------------------------+
|      message                                                         |
+----------------------------------------------------------------------*/

static char startup_message[] =
"\n"
"This is '" MYNAME " " MYVERSION "' (A Promising Chess Program)\n"
"Copyright (C)1981-2006 Yves Debilloez\n"
"Please refer to the license terms (license.html) before using.\n"
"\n"
"Type 'help' for a list of commands.\n"
"Type 'protover 4' to have more modern interface." ;

/*----------------------------------------------------------------------+
|      commands                                                        |
+----------------------------------------------------------------------*/

static void	cmd_not_supported(char * szCommand);
static void	cmd_setboard(char * szCommand,char * szArg);
static void cmd_signalcatch(int s);
static void cmd_bd(UNUSED char * szCommand,UNUSED char * szArg);
static void cmd_save(UNUSED char * szCommand,UNUSED char * szArg);

static void cmd_snd_protover(int nProtover);
static void cmd_snd_error(char *szCommand,char *szReason);
static void cmd_snd_errormove(char *szMove);

static void cmd_about(char * szCommand,char * szArg);

static int cmd_readarg(char * szLine, char * szArg,const size_t nSizeArg);
static int cmd_readchar(void *pFile);

static void stat_logadd(const char *szLine,const t_boolean nDir);

void cmd_quit(char * szCommand,char * szArg);
void cmd_force(char * szCommand,char * szArg);

/**
 * configure the engine and all local parameters
 * this is called when a total reset is required such as launching or
 * relaunching the application whilst in memory
 */
void cmd_constructor(UNUSED char * szCommand,UNUSED char * szArg)
{
	// make sure we catch this signal
	signal(SIGINT, cmd_signalcatch);

	post_mode = 1;
	interface_mode = 1; // assume we only support level 1 commands unless told differently
	ics_mode = 0;
	processing_mode = 0;
	
	setbuf(stdout,NULL);
	setbuf(stdin,NULL);
	
	if (!xboard_mode)
		cmd_about("","");
	
	cmd_new("","nosave");  // further initialisation can be done in new, beware for sideeffects
}

/**
 * second constructor, taking more time and issued during protover command
 */
void cmd_constructor2(void)
{
	static int iInit = myFALSE;
	
	if (!iInit) {
		// make sure it is only executed once
		iInit = myTRUE;
		book_load(BOOKNAME);
	}
}

// --------------------------------------------------------------------

static void cmd_signalcatch(int s)
{
	signal(s, cmd_signalcatch);
}

/**
 * gnuchess compatible implementation
 */
void cmd_new(UNUSED char * szCommand,UNUSED char * szArg)
{
	random_mode = 0;
	force_mode = 0;
	ponder_mode = 0;
	
	if (strcmp(szArg,"nosave"))
		cmd_save("","");

	computer_engines[0] = 0;
	computer_engines[1] = 1;
	
	curply = 0;
	strcpy(szResult,"*");
	strcpy(szOpponentName,"?");
	strcpy(szSite,"127.0.0.1");

	cmd_setboard("","rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
	if ((!xboard_mode) && (!processing_mode))
		cmd_bd("","");
}

/**
 * print the board to the user
 */
static void cmd_bd(UNUSED char * szCommand,UNUSED char * szArg)
{
	board_print();
}

/**
 * set xboard mode, xboard means compatible
 */
static void cmd_xboard(UNUSED char * szCommand,UNUSED char * szArg)
{
	xboard_mode = 1;
	cmd_snd("");
}

/**
 * toggle random mode
 */
static void cmd_random(UNUSED char * szCommand,UNUSED char * szArg)
{
	random_mode = 1 - random_mode;
}

/**
 * allow to set game variation, is not supported
 */
static void cmd_variant(char * szCommand,UNUSED char * szArg)
{
	cmd_not_supported(szCommand);
}

/** 
 * exit engine
 * also understands bye commands
 */
void cmd_quit(UNUSED char * szCommand,UNUSED char * szArg)
{
	exit(0);
}

/**
 * computer plays white
 */
static void cmd_white(UNUSED char * szCommand,UNUSED char * szArg)
{
	computer_engines[0] = 1;
	computer_engines[1] = 0;
	// TODO: check if need to start moving
}

/**
 * computer plays black
 */
static void cmd_black(UNUSED char * szCommand,UNUSED char * szArg)
{
	computer_engines[0] = 0;
	computer_engines[1] = 1;
	// TODO: check if need to start moving
}

/** 
 * switch to force mode, turn off ponder
 */
void cmd_force(UNUSED char * szCommand,UNUSED char * szArg)
{
	force_mode = 1;
	thinking = 0;
}

/**
 * computer plays colour to move
 */
void cmd_go(UNUSED char * szCommand,UNUSED char * szArg)
{
	if (curply % 2) cmd_black("","");
	else cmd_white("","");
	force_mode = 0;
}

/**
 * play oponent colour, start pondering, and make sure to start
 * thinking as soon as the last move is received
 * (Freechess sends playother and then the last move)
 */
static void cmd_playother(UNUSED char * szCommand,UNUSED char * szArg)
{
	if (curply % 2) cmd_white("","");
	else cmd_black("","");
	ponder_mode = 1;
    force_mode = 0;
}

/**
 * configure board using fen string
 */
static void cmd_setboard(UNUSED char * szCommand,char * szArg)
{
	board_setup(szArg);
}

/**
 * move now, registered as question mark
 */
static void cmd_movenow(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_force("","");
}

static void cmd_protover(UNUSED char * szCommand,char * szArg)
{
	sscanf(szArg, "%d", &interface_mode);
	
	cmd_snd_protover(interface_mode);
}

static void cmd_level(UNUSED char * szCommand,char * szArg)
{
	level_mode = 1;
	sscanf(szArg, "%d", &level);
}

void cmd_sd(UNUSED char * szCommand,char * szArg)
{
	level_mode = 0;
	sscanf(szArg, "%f", &maxdepth);
}

static void cmd_st(UNUSED char * szCommand,UNUSED char * szArg)
{
}

static void cmd_time(UNUSED char * szCommand,UNUSED char * szArg)
{
	sscanf(szArg,"%d",&time_left);
	
	// convert from centiseconds to seconds
	time_left = time_left / 100;
}

static void cmd_otim(UNUSED char * szCommand,UNUSED char * szArg)
{
	char szResult[MAX_DATANAME_LENGTH];
	
	sprintf(szResult,"got otim %s",szArg);
	cmd_snd_info(szResult);
}

/**
 * give a hint about the best move, if no hint available, ignore
 */
static void cmd_hint(UNUSED char * szCommand,UNUSED char * szArg)
{
}

static void cmd_draw(UNUSED char * szCommand,UNUSED char * szArg)
{
}

static void cmd_computer(UNUSED char * szCommand,UNUSED char * szArg)
{
}

static void cmd_exit(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_not_supported("exit");
}
static void cmd_pause(UNUSED char * szCommand,UNUSED char * szArg)
{
}

static void cmd_resume(UNUSED char * szCommand,UNUSED char * szArg)
{
}

static void cmd_result(UNUSED char * szCommand,UNUSED char * szArg)
{
	// cut out comments from standard results, keep only the score data normalized
	if (strncmp("1-0",szArg,3)==0) {
		strcpy(szResult,"1-0");
	} else if (strncmp("0-1",szArg,3)==0) {
		strcpy(szResult,"0-1");
	} else if (strncmp("1/2-1/2",szArg,7)==0) {
		strcpy(szResult,"1/2-1/2");
	} else {
		// in other cases, just take the string
		strcpy(szResult,szArg);
	}
	
	cmd_save("","");
	force_mode = 1;
}

/**
 * create a book from a PGN file
 */
static void cmd_createbook(UNUSED char * szCommand,char * szArg)
{
	book_create(szArg,BOOKNAME);
}

/**
 * load a different book from a book file
 */
static void cmd_loadbook(UNUSED char * szCommand, char *szArg)
{
    book_load(szArg);
}

/**
 * recieve message from xboard that we are running on ICS server
 */
static void cmd_ics(UNUSED char * szCommand,char * szArg)
{
	if (strcmp(szArg,"-")!=0) {
		ics_mode = 1;
		strcpy(szSite,szArg);
	}
}

/**
 * tell something about myself
 */
static void cmd_about(UNUSED char * szCommand,UNUSED char * szArg)
{
	puts(startup_message);
}

static void cmd_edit(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_not_supported("edit");
}

static void cmd_bk(UNUSED char * szCommand,UNUSED char * szArg)
{
	book_listmoves(m_boardhistory[curply].szFEN,myTRUE);
}

/**
 * undo one move
 */
static void cmd_undo(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_force("","");
	board_unapplymove();
	board_print();
}

/**
 * undo the move and play it again
 */
static void cmd_again(UNUSED char * szCommand, UNUSED char *szArg)
{
	cmd_force("","");
	board_unapplymove();
	cmd_go("","");
}

/**
 *
 */
static void cmd_remove(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_force("","");
	board_unapplymove();
	board_unapplymove();
	board_print();

	// TODO: now make sure to start thinking again
}

/**
 * save current game to a file
 */
static void stat_savefile(char * szFileName, const t_boolean bAppend)
{
	FILE *fpSave;
	int i;
	int iSpace=myTRUE;
	
	struct tm *p_tm;
    time_t l_tm;
    char szTime[80];
    
	if (curply>0) {
		l_tm = time(NULL);			// get system time
		p_tm = localtime(&l_tm);	// return local time in the form of tm structure    
		strftime(szTime,sizeof(szTime),"%Y.%m.%d",p_tm);
		
        fpSave = fopen(szFileName,bAppend==myTRUE?"a+":"w+");
		if (fpSave!=NULL) {
			fprintf(fpSave,"[Event \"%s autosave\"]\n",MYNAME);
			fprintf(fpSave,"[Site \"%s\"]\n",szSite);
			fprintf(fpSave,"[Date \"%s\"]\n",szTime);
			fprintf(fpSave,"[Round \"?\"]\n");
			fprintf(fpSave,"[White \"%s\"]\n",computer_engines[0]?MYNAME " " MYVERSION:szOpponentName);
			fprintf(fpSave,"[Black \"%s\"]\n",computer_engines[1]?MYNAME " " MYVERSION:szOpponentName);
			fprintf(fpSave,"[Result \"%s\"]\n",szResult);
			fprintf(fpSave,"\n");
			for (i=0;i<min(curply,MAX_GAME_LENGTH-1);i++) {
				if ((i%2)==0) {
					fprintf(fpSave,"%d. %s",(i/2)+1,m_boardhistory[i].szSANMove);
					iSpace = myFALSE;
				} else { 
					fprintf(fpSave," %s ",m_boardhistory[i].szSANMove);
					iSpace = myTRUE;
				}
				if (((i)%10)==9) {
					fprintf(fpSave,"\n");
					iSpace = myTRUE;
				}
			}
			if (!iSpace) fprintf(fpSave," ");
			fprintf(fpSave,"%s\n\n",szResult);
			fclose(fpSave);
		}
	}
}

/**
 * save the game to a file
 */
static void cmd_save(char * szCommand,UNUSED char * szArg)
{
    if (strlen(szCommand)==0) {
        stat_savefile(PGNNAME, myTRUE);
    } else {
        stat_savefile(szCommand, myFALSE);
    }
}

/**
 * load an EPD file and set position to it
 * TODO: then interprete commands and act on them
 */
static void cmd_epd(UNUSED char * szCommand,UNUSED char * szArg)
{
 	FILE *fpOpen;
    char szBuffer[MAX_DATANAME_LENGTH];
    
    fpOpen = fopen(szCommand,"r");
    if (fpOpen != NULL) {
        file_readline(szBuffer,fpOpen,sizeof(szBuffer));
        fclose(fpOpen);
        
        cmd_setboard(szBuffer,"");
    }
}

/**
 * load a file and execute all moves
 */
static void cmd_load(UNUSED char * szCommand,UNUSED char * szArg)
{
    cmd_not_supported("load");
}

/**
 * go into uci mode
 */
static void cmd_uci(UNUSED char * szCommand,UNUSED char * szArg)
{
    cmd_not_supported("uci");
}

/**
 * start pondering
 */
static void cmd_hard(UNUSED char * szCommand,UNUSED char * szArg)
{
	ponder_mode = 1;
}

/**
 * stop pondering
 */
static void cmd_easy(UNUSED char * szCommand,UNUSED char * szArg)
{
	ponder_mode = 0;
}

static void cmd_post(UNUSED char * szCommand,UNUSED char * szArg)
{
	post_mode = 1;
}

static void cmd_nopost(UNUSED char * szCommand,UNUSED char * szArg)
{
	post_mode = 0;
}

static void cmd_ping(UNUSED char * szCommand,char * szArg)
{
	char szResult[MAX_DATANAME_LENGTH];
	
	sprintf(szResult,"pong %s",szArg);
	cmd_snd(szResult);
}

static void cmd_analyze(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_not_supported("analyze");
}

/**
 * print game to console
 */
static void cmd_game(UNUSED char * szCommand,UNUSED char * szArg)
{
	int i;
    char szBuffer[MAX_DATANAME_LENGTH];
    
	for (i=0;i<min(curply,MAX_GAME_LENGTH-1);i++) {
		if ((i%2)==0)
			printf(" %d. %s",(i/2)+1,m_boardhistory[i].szSANMove);
		else {
			sprintf(szBuffer," %s ",m_boardhistory[i].szSANMove);
            cmd_snd(szBuffer);
        }
	}
	if (min(curply,MAX_GAME_LENGTH-1)%2)
		cmd_snd("");
}

static void cmd_name(UNUSED char * szCommand,UNUSED char * szArg)
{
	char szResult[MAX_DATANAME_LENGTH];

	strcpy(szOpponentName,szArg); // used when saving game

	if (ics_mode) {
		sprintf(szResult,"tellopponent Welcome to my domain %s!  I am a computer.",szArg);
		cmd_snd(szResult);
		sprintf(szResult,"tellopponent My version is %s, finger %s for more details",MYVERSION,MYNAME);
		cmd_snd(szResult);
	}
}

static void cmd_rating(UNUSED char * szCommand,UNUSED char * szArg)
{
	char szResult[MAX_DATANAME_LENGTH];
	
	sprintf(szResult,"got rating %s",szArg);
	cmd_snd_info(szResult);
}

/**
 * return position evaluation
 */
static void cmd_eval(UNUSED char * szCommand,UNUSED char * szArg)
{
	t_score fResult;
	
	fResult = eval_pos(curply & 1);
	cmd_snd_thinking(0,fResult,"evaluation");
}

/**
 * TODO: set flags for all accepted features
 */
static void cmd_accepted(UNUSED char * szCommand,UNUSED char * szArg)
{
}

/**
 * TODO: handle the fact that there is an incompatibility
 */
static void cmd_rejected(UNUSED char * szCommand,UNUSED char * szArg)
{
	cmd_not_supported("rejected");
}

/**
 * new version of move input
 */
static void cmd_usermove(UNUSED char * szCommand,char * szArg)
{
	if (board_applymove(szArg)==MOVERESULT_ERROR)
		cmd_snd_errormove(szArg);
}

/**
 * play both colours
 */
static void cmd_both(UNUSED char * szCommand,UNUSED char * szArg)
{
	computer_engines[0] = 1;
	computer_engines[1] = 1;
	force_mode = 0;
}

/**
 * select the engine to be used
 */
static void cmd_alg(UNUSED char * szCommand,char * szArg)
{
	sscanf(szArg, "%d", &algoritme);
}

// --------------------------------------------------------------------

static void cmd_coordmove(char * szCommand,UNUSED char * szArg)
{
	// as of XBoard 4.2.7 force the use of usermode
	if (interface_mode==2) { 
		cmd_not_supported(szCommand);
	} else {
		if (board_applymove(szCommand)==MOVERESULT_ERROR    )
			cmd_snd_errormove(szArg);
	}	
}

// --------------------------------------------------------------------

static void cmd_not_supported(char *szCommand)
{
	if (!xboard_mode)
		cmd_snd_error(szCommand,"unknown command");
}

/**
 * reply to xboard, main and only routine for this purpose
 */
void cmd_snd(const char *szCommand)
{
    // make sure not to add the newline twice
    if ((strlen(szCommand)>0) && (szCommand[strlen(szCommand)-1]=='\n')) {
        printf(szCommand);
    } else {
        printf("%s\n",szCommand);
    }
    stat_logadd(szCommand,myTRUE);
}

/**
 * send the move generated by the engine to xboard
 */
void cmd_snd_move(const char *szMove)
{
    char szBuffer[MAX_DATANAME_LENGTH];  
	int iResult;
	
	iResult = board_applymove(szMove);
	if (iResult == 0) {
		sprintf(szBuffer,"move %s",szMove); cmd_snd(szBuffer);
	} else if (iResult == 1) {
		// one move, draw by lack of material
		sprintf(szBuffer,"move %s",szMove); cmd_snd(szBuffer);
		cmd_snd("1/2-1/2 {Draw by lack of material}");
		strcpy(szResult,"1/2-1/2");
	} else if (iResult == 2) {
		// one move, draw by threefold repetition
		sprintf(szBuffer,"move %s",szMove); cmd_snd(szBuffer);
		cmd_snd("1/2-1/2 {Draw by threefold repetition}");
		strcpy(szResult,"1/2-1/2");
	} else if (iResult == 3) {
		// one move, draw by 50 move rule
		sprintf(szBuffer,"move %s",szMove); cmd_snd(szBuffer);
		cmd_snd("1/2-1/2 {Draw by 50 move rule}");
		strcpy(szResult,"1/2-1/2");
	} else if (iResult == -1) {
		// one move, draw by stalemate
		cmd_snd("1/2-1/2 {Stalemate}");
		strcpy(szResult,"1/2-1/2");
	} else if (iResult == -2) {
		// one move, black wins
		cmd_snd("0-1 {Black wins}");
		strcpy(szResult,"0-1");
	} else if (iResult == -3) {
		// one move, white wins
		cmd_snd("1-0 {White wins}");
		strcpy(szResult,"1-0");
	} else if (iResult == -4) {
		// one move, black resigns
		cmd_snd("1-0 {Black resigns}");
		strcpy(szResult,"1-0");
	} else if (iResult == -5) {
		// one move, white resigns
		cmd_snd("0-1 {White resigns}");
		strcpy(szResult,"0-1");
	}
	
	// make sure that after end of game, searching ends
	if (iResult != 0) {
		cmd_force("","");
		cmd_save("","");
	}
}

/**
 * return illegal move back to xboard
 */
static void cmd_snd_errormove(char *szMove)
{
    char szBuffer[MAX_DATANAME_LENGTH];    
	sprintf(szBuffer,"Illegal move: %s",szMove);
    cmd_snd(szBuffer);
}

/**
 * return an error back to xboard
 */
static void cmd_snd_error(char *szCommand,char *szReason)
{
    char szBuffer[MAX_DATANAME_LENGTH];    
	sprintf(szBuffer, "Error (%s): %s", szReason, szCommand);    
    cmd_snd(szBuffer);
}

/**
 * return an info back to xboard
 */
void cmd_snd_info(const char *szMessage)
{
    char szBuffer[MAX_DATANAME_LENGTH];
    
	if (post_mode) {
        sprintf(szBuffer," %s",szMessage);
        cmd_snd(szBuffer);
    }
}

/**
 * return thinking statistics back
 */
void cmd_snd_thinking(const int nPly, const t_score fEval, const char *szThinking)
{
    char szBuffer[MAX_DATANAME_LENGTH];
    
	if (post_mode) {
        if (nPly != 0) {
            sprintf(szBuffer," %3d %2.3f %s",nPly,fEval,szThinking);
        } else {
            sprintf(szBuffer," %2.3f %s",fEval,szThinking);
        }
        cmd_snd_info(szBuffer);
    }
}

/**
 *
 */
static void cmd_snd_protover(UNUSED int nProtover)
{
	cmd_snd("feature done=0");
	
	// TODO: replace by code that will interprete the command structure and disable
	// features dynamically
	cmd_snd("feature ping=1 setboard=1 playother=1 san=0 usermove=1 time=1");
	cmd_snd("feature draw=0 sigint=1 sigterm=1 reuse=1");
	cmd_snd("feature ics=1 name=1 pause=0 analyze=0 colors=0");

	cmd_snd("feature myname=\"" MYNAME " " MYVERSION "\"");
	cmd_snd("feature variants=\"normal\"");
	
	// time intensive constructor
	cmd_constructor2();
	
	cmd_snd("feature done=1");
}

// --------------------------------------------------------------------
#define CMD_NOPARAM				0
#define CMD_OPT					1
#define CMD_PARAM_STRING		2
#define CMD_PARAM_SWITCH		4
#define CMD_PARAM_NUMBER		8
#define CMD_PARAM_NOCHECK		16

/**
 * show help on screen, information shown depends on protocol level
 */
static void cmd_help(UNUSED char * szCommand,char * szArg)
{
	struct st_command *c;
	
	if (strlen(szArg)) {
		// request more details on specific command
		c = list_commands;
		do {
			if ((c->cmd_name != NULL)&&(!strcmp(szArg,c->cmd_name))) {
				// command found
				printf("%-10s - %s\n", c->cmd_name ? c->cmd_name : "", c->cmd_help);
				if (c->cmd_paramdata == CMD_NOPARAM )
					printf("\tthe function does not take any parameters\n");
				else if (c->cmd_paramdata & CMD_OPT)
					printf("\tthe function does take optional parameters\n");
				if (c->cmd_paramdata & CMD_PARAM_STRING )
					printf("\t%s string",c->cmd_name ? c->cmd_name : "");
				else if (c->cmd_paramdata & CMD_PARAM_NUMBER )
					printf("\t%s n",c->cmd_name ? c->cmd_name : "");
				else if (c->cmd_paramdata & CMD_PARAM_SWITCH )
					printf("\t%s on|off (default is on)",c->cmd_name ? c->cmd_name : "");
			}
		} while (c++->cmd_name != NULL);
	} else {
		// print all commands
		cmd_snd("commands are:");
		c = list_commands;
		do {
			if (c->cmd_interfacelevel & interface_mode) {
				// is this command supported in this interface mode ?
				printf("%-10s - %s\n", c->cmd_name ? c->cmd_name : "", c->cmd_help);
			}
		} while (c++->cmd_name != NULL);
	}
	cmd_snd("");
}

/**
 * list of st_command pointers
 */
static struct st_command list_commands[] = {
	{ "help",		cmd_help,		"show this list of commands"			, 7, CMD_PARAM_STRING | CMD_OPT },
	{ "?",			cmd_movenow,	"force the computer to move now"		, 1, CMD_NOPARAM },
	{ "about",		cmd_about,		"tell something about myself"			, 4, CMD_NOPARAM },
	{ "accepted",	cmd_accepted,	"a requested feature is implemented"	, 4, CMD_PARAM_STRING },
	{ "again",		cmd_again,		"research again from this position"		, 4, CMD_NOPARAM },
	{ "alg",		cmd_alg,		"select algorithm"						, 2, CMD_PARAM_NUMBER },
	{ "analyze",	cmd_analyze,	"enter analyze mode"					, 1, CMD_PARAM_STRING | CMD_OPT },
	{ "bd",			cmd_bd,			"display board"							, 4, CMD_NOPARAM },
	{ "bk",			cmd_bk,			"book"									, 1, CMD_PARAM_NUMBER | CMD_OPT },
	{ "black",		cmd_black,		"set computer to play black"            , 1, CMD_NOPARAM },
	{ "both",		cmd_both,		"set computer to play both"             , 7, CMD_NOPARAM },
	{ "computer",	cmd_computer,	"playing against a computer"			, 2, CMD_NOPARAM },
	{ "createbook",	cmd_createbook,	"create the opening book from PGN"		, 4, CMD_PARAM_STRING },
	{ "draw",		cmd_draw,		"offer draw"							, 1, CMD_NOPARAM },
	{ "easy",		cmd_easy,		"do not think when not on move"         , 1, CMD_NOPARAM },
	{ "edit",		cmd_edit,		"go to edit mode"						, 1, CMD_PARAM_STRING | CMD_OPT },
    { "epd",        cmd_epd,        "load a game in EPD form from disk"     , 4, CMD_PARAM_STRING},
	{ "eval",		cmd_eval,		"give evaluation of function"			, 4, CMD_NOPARAM },
	{ "exit",		cmd_exit,		"leave analysis mode"                   , 4, CMD_NOPARAM },
	{ "force",		cmd_force,		"computer plays neither side"           , 7, CMD_NOPARAM },
	{ "game",		cmd_game,		"print current game"					, 4, CMD_NOPARAM },
	{ "go",			cmd_go,			"computer starts playing"               , 7, CMD_NOPARAM },
	{ "hard",		cmd_hard,		"computer thinks all the time"          , 1, CMD_NOPARAM },
	{ "hint",		cmd_hint,		"computer gives a hint"					, 1, CMD_NOPARAM },
	{ "ics",		cmd_ics,		"set hostname if playing on ics"		, 2, CMD_PARAM_STRING },
	{ "level",		cmd_level,		"select level to play at"               , 1, CMD_PARAM_NUMBER },
	{ "load",		cmd_load,		"load game in pgn form from disk"       , 4, CMD_PARAM_STRING },
	{ "loadbook",	cmd_loadbook,	"load book from disk"                   , 4, CMD_PARAM_STRING },
	{ "name",		cmd_name,		"give your name"						, 1, CMD_PARAM_STRING },
	{ "new",		cmd_new,		"new game"                              , 7, CMD_NOPARAM },
	{ "nopost",		cmd_nopost,		"stop outputting search details"		, 3, CMD_NOPARAM },
	{ "otim",		cmd_otim,		"set opponent time"						, 1, CMD_PARAM_NUMBER },
	{ "pause",		cmd_pause,		"suspend"								, 2, CMD_NOPARAM },
	{ "ping",		cmd_ping,		"check if still alive"					, 2, CMD_NOPARAM },
	{ "playother",	cmd_playother,	"set computer to play the other colour" , 2, CMD_NOPARAM },
	{ "post",		cmd_post,		"output search details"					, 7, CMD_PARAM_SWITCH },
	{ "protover",	cmd_protover,	"set command interface level"			, 7, CMD_PARAM_NUMBER },
	{ "quit",		cmd_quit,		"leave chess program"                   , 7, CMD_NOPARAM },
	{ "random",		cmd_random,		"causes the computer to play randomly"  , 1, CMD_PARAM_SWITCH | CMD_OPT },
	{ "rating",		cmd_rating,		"set computer and player rating"		, 1, CMD_PARAM_NOCHECK },
	{ "rejected",	cmd_rejected,	"a feature is not supported"			, 2, CMD_PARAM_STRING },
	{ "remove",		cmd_remove,		"undo last 2 moves"						, 7, CMD_PARAM_NUMBER | CMD_OPT },
	{ "result",		cmd_result,		"set result string"						, 7, CMD_PARAM_STRING },
	{ "resume",		cmd_resume,		"unpause"								, 2, CMD_PARAM_STRING | CMD_OPT },
	{ "save",		cmd_save,		"save game in PGN format"				, 4, CMD_NOPARAM },
	{ "sd",			cmd_sd,			"select depth to play at"               , 1, CMD_PARAM_NUMBER },
	{ "st",			cmd_st,			"set time controls"						, 7, CMD_PARAM_NOCHECK },
	{ "time",		cmd_time,		"set own clock"							, 1, CMD_PARAM_NUMBER },
	{ "setboard",   cmd_setboard,	"setup new position"                    , 6, CMD_PARAM_STRING },
	{ "uci",		cmd_uci,		"enter uci mode"                        , 1, CMD_NOPARAM },
	{ "undo",		cmd_undo,		"undo move"                             , 1, CMD_NOPARAM },
	{ "usermove",	cmd_usermove,	"enter moves in algebraic notation"		, 2, CMD_PARAM_STRING },
	{ "variant",	cmd_variant,	"set chess variant"						, 7, CMD_PARAM_STRING },
	{ "white",		cmd_white,		"set computer to play white"            , 1, CMD_NOPARAM },
	{ "xboard",		cmd_xboard,		"switch to xboard mode"                 , 7, CMD_PARAM_SWITCH | CMD_OPT },
	{ NULL,			cmd_coordmove,	"enter moves in coordinate notation"    , 1, CMD_NOPARAM },
};

/**
 * return true when there is a command in the queue
 */
int cmd_peek(void)
{
#if defined(_MSC_VER)
	return 0;
#else
    static fd_set   readfds;
    struct timeval  peektimeout;
    int             iRet;
	int				nSelect;
	
    FD_ZERO(&readfds);
    FD_SET(fileno(stdin), &readfds);

    peektimeout.tv_sec = 0;
    peektimeout.tv_usec = 0;
    nSelect = select(fileno(stdin)+1, &readfds, NULL, NULL, &peektimeout);
	if (nSelect==0) 
		// no descriptors ready
		return 0;
    
	iRet = FD_ISSET(fileno(stdin), &readfds);
    
	return (iRet);
#endif
}

/**
 * read input command from xboard
 * return codes not used now
 */
int cmd_readcommand(void)
{
	char szArg[MAX_DATANAME_LENGTH];
	char szCommand[MAX_DATANAME_LENGTH];
	int  iCmd;
	
	for(;;)
	{
		if (!xboard_mode) {
			fputs("Belofte> ", stdout);
		}
		// fflush(stdout);  // not required since there is a setbuf	
		cmd_readline(szCommand,stdin,sizeof(szCommand));
		cmd_readarg(szCommand,szArg,sizeof(szArg));
        
        stat_logadd(szCommand,myFALSE);
		
		for (iCmd=0; list_commands[iCmd].cmd_name != NULL; iCmd++) {
            // compare to all except last one
			if (0==strcmp(list_commands[iCmd].cmd_name, szCommand)) {
				break;
			}
		}
        list_commands[iCmd].cmd_ptr(szCommand,szArg);
        
		break;
	}
	
	return 0;
}

/**
 * read line
 * @return 0 if there is an end of file
 */
int cmd_readline(char * szBuffer, void *pFile, const size_t nSize)
{
	unsigned int iIndex=0;
	int iCurrent;
	unsigned char cCurrent;
    int nCommentLevel=0;
    int nReadChars = 0;
	
	if (nSize<2) return 0;
	szBuffer[iIndex] = 0;
	
	while(!feof((FILE *)pFile)) {
		iCurrent = cmd_readchar(pFile);
		if (iCurrent == -1) break;
		cCurrent = (unsigned char) iCurrent;
        nReadChars++;
		if ((cCurrent == '\n')||(cCurrent == '\r')) {
            if (nCommentLevel == 0) {
                break;
            }
        } else if (cCurrent == '{') {
            nCommentLevel ++;
        } else if (cCurrent == '}') {
            nCommentLevel --;
        } else if ((nCommentLevel > 0) && (cCurrent == ']')) {
            // break out of error reading
            nCommentLevel = 0;
		} else {
            if (nCommentLevel == 0) {
                szBuffer[iIndex++] = cCurrent;
                szBuffer[iIndex] = 0;
            }
		}
		if (iIndex+1>=nSize) break;
        if (nReadChars>5000) {
            stat_logadd("Reading more than 5000 chars",myTRUE);
            return -1; // make sure to break out if comment malformed
        }
	}
    if (feof((FILE *)pFile)) return -1;
    
	return 0;
}

/**
* read line
 * @return 0 if there is an end of file
 */
int file_readline(char * szBuffer, FILE *pFile, const size_t nSize)
{
	unsigned int iIndex=0;
	int iCurrent;
	unsigned char cCurrent;
    int nCommentLevel=0;
    int nReadChars = 0;
	
	if (nSize<2) return 0;
	szBuffer[iIndex] = 0;
	
	while(!feof(pFile)) {
		iCurrent = fgetc(pFile);
        if (iCurrent == EOF) break;
        
		cCurrent = (unsigned char) iCurrent;
        nReadChars++;
		if (cCurrent == '\n') {
            if (nCommentLevel == 0) {
                break;
            }
        } else if (cCurrent == '\r') {
            // we skip this character
            // for Mac OS 9 and prior files this will fail
            // on Mac OX X and DOS/UNIX, no problem
        } else if (cCurrent == '{') {
            nCommentLevel ++;
        } else if (cCurrent == '}') {
            nCommentLevel --;
        } else if ((nCommentLevel > 0) && (cCurrent == ']')) {
            // break out of error reading
            nCommentLevel = 0;
		} else {
            if (nCommentLevel == 0) {
                szBuffer[iIndex++] = cCurrent;
                szBuffer[iIndex] = 0;
            }
		}
		if (iIndex+1>=nSize) break;
        if (nReadChars>5000) {
            stat_logadd("Reading more than 5000 chars",myTRUE);
            return -1; // make sure to break out if comment malformed
        }
	}
    if (feof(pFile)) return -1;
    
	return 0;
}

/**
 * read one character at a time
 */
static int cmd_readchar(void *pFile)
{
	unsigned char aBuf[2];
	
    // break on end of file
    if (feof((FILE *)pFile)) return -1;
    
	read(fileno(pFile),aBuf,1);
	
	return (int) aBuf[0];
}

/**
 * split arguments on line
 */
static int cmd_readarg(char * szLine, char * szArg,const size_t nSizeArg)
{
	int i;
	int nSizeLine = strlen(szLine);
	
	if (nSizeArg<1) return 0;
		
	szArg[0] = '\0';
	
	for (i=0;i<nSizeLine;i++) {
		if (szLine[i] == ' ') {
			strncpy(szArg,szLine+i+1,nSizeArg);
			szArg[nSizeArg-1] = '\0';
			szLine[i] = '\0';
			// TODO: left trim arguments so there are no leading spaces
			break;
		}
	}
	return 1;
}

// --------------------------------------------------------------------

static void stat_logadd(const char *szLine,const t_boolean nDir)
{
    FILE *fpLog;
	
	struct tm *p_tm;
    time_t l_tm;
    char szTime[80];
    
	l_tm = time(NULL);			// get system time
	p_tm = localtime(&l_tm);	// return local time in the form of tm structure    
	strftime(szTime,sizeof(szTime),"%Y.%m.%d",p_tm);
		
    fpLog = fopen(LOGNAME,"a+");
    if (fpLog!=NULL) {
		fprintf(fpLog,"%s %s %s\n",nDir==myTRUE?"send":"receive",szTime,szLine);
        fclose(fpLog);
	}
}

// --------------------------------------------------------------------
// eof